import unittest

from riptable import *
from riptable.rt_datetime import NANOS_PER_DAY, YDAY_SPLITS, YDAY_SPLITS_LEAP

numeric_dt = [
    np.int8,
    np.uint8,
    np.int16,
    np.uint16,
    np.int32,
    np.uint32,
    np.int64,
    np.uint64,
    np.float32,
    np.float64,
]


class Date_Test(unittest.TestCase):
    def test_constructor_numeric(self):
        # python scalars
        d = Date(1)
        self.assertTrue(isinstance(d, Date))
        self.assertEqual(d._fa[0], 1)
        self.assertEqual(d.day_of_year[0], 2)
        self.assertEqual(d.year[0], 1970)
        self.assertEqual(d.month[0], 1)

        d = Date(1.0)
        self.assertTrue(isinstance(d, Date))
        self.assertEqual(d._fa[0], 1)
        self.assertEqual(d.day_of_year[0], 2)
        self.assertEqual(d.year[0], 1970)
        self.assertEqual(d.month[0], 1)

        # python lists
        d = Date([1])
        self.assertTrue(isinstance(d, Date))
        self.assertEqual(d._fa[0], 1)
        self.assertEqual(d.day_of_year[0], 2)
        self.assertEqual(d.year[0], 1970)
        self.assertEqual(d.month[0], 1)

        d = Date([1.0])
        self.assertTrue(isinstance(d, Date))
        self.assertEqual(d._fa[0], 1)
        self.assertEqual(d.day_of_year[0], 2)
        self.assertEqual(d.year[0], 1970)
        self.assertEqual(d.month[0], 1)

        # numeric numpy arrays
        for dt in numeric_dt:
            d = Date(FA(1, dtype=dt))
            self.assertEqual(d._fa[0], 1)
            self.assertEqual(d.day_of_year[0], 2)
            self.assertEqual(d.year[0], 1970)
            self.assertEqual(d.month[0], 1)

    def test_constructor_string(self):
        yeardays = [46, 365, 158, 331]
        months = [2, 12, 6, 11]
        years = [2016, 2017, 2018, 2099]
        d = Date(["2016-02-15", "2017-12-31", "2018-06-07", "2099-11-27"])
        self.assertTrue(isinstance(d, Date))
        self.assertTrue(bool(np.all(yeardays == d.day_of_year)))
        self.assertTrue(bool(np.all(months == d.month)))
        self.assertTrue(bool(np.all(years == d.year)))

    def test_constructor_format(self):
        yeardays = [46, 365, 158, 331]
        months = [2, 12, 6, 11]
        years = [2016, 2017, 2018, 2099]
        d = Date(["02_15_2016", "12_31_2017", "06_07_2018", "11_27_2099"], format="%m_%d_%Y")
        self.assertTrue(isinstance(d, Date))
        self.assertTrue(bool(np.all(yeardays == d.day_of_year)))
        self.assertTrue(bool(np.all(months == d.month)))
        self.assertTrue(bool(np.all(years == d.year)))

    def test_constructor_categorical(self):
        yeardays = [46, 365, 158, 331]
        months = [2, 12, 6, 11]
        years = [2016, 2017, 2018, 2099]
        c = Categorical(["2016-02-15", "2017-12-31", "2018-06-07", "2099-11-27"])
        d = Date(c)
        self.assertTrue(isinstance(d, Date))
        self.assertTrue(bool(np.all(yeardays == d.day_of_year)))
        self.assertTrue(bool(np.all(months == d.month)))
        self.assertTrue(bool(np.all(years == d.year)))

    def test_constructor_nano(self):
        pass

    def test_constructor_matlab_datenum(self):
        yeardays = arange(1, 6)
        months = 1
        years = 2018
        d = FA([737061.0, 737062.0, 737063.0, 737064.0, 737065.0])
        d = Date(d, from_matlab=True)
        self.assertTrue(isinstance(d, Date))
        self.assertTrue(bool(np.all(yeardays == d.day_of_year)))
        self.assertTrue(bool(np.all(months == d.month)))
        self.assertTrue(bool(np.all(years == d.year)))

    def test_constructor_empty(self):
        d = Date([])
        self.assertTrue(len(d) == 0)
        self.assertTrue(isinstance(d, Date))
        self.assertTrue(d.dtype == np.int32)

    def test_subtract(self):
        # TODO: add more assertions
        # only return type is checked now
        # most combinations of not allowed types will be caught by FastArray

        d = Date(10)
        val = 1
        result = d - val
        self.assertTrue(isinstance(result, Date))

        val = [1]
        result = d - val
        self.assertTrue(isinstance(result, Date))

        val = FastArray(1)
        result = d - val
        self.assertTrue(isinstance(result, Date))

        val = Date(1)
        result = d - val
        self.assertTrue(isinstance(result, DateSpan))

        val = DateSpan(1)
        result = d - val
        self.assertTrue(isinstance(result, Date))

        val = DateTimeNano(["1970-01-10"], from_tz="GMT")
        result = d - val
        self.assertTrue(isinstance(result, TimeSpan))

        val = TimeSpan(1, unit="h")
        result = d - val
        self.assertTrue(isinstance(result, DateTimeNano))

    def test_add(self):
        d = Date(10)
        val = 1
        result = d + val
        self.assertTrue(isinstance(result, Date))

        val = 1.0
        result = d + val
        self.assertTrue(isinstance(result, Date))

        val = [1]
        result = d + val
        self.assertTrue(isinstance(result, Date))

        val = FastArray(1)
        result = d + val
        self.assertTrue(isinstance(result, Date))

        val = DateSpan(1)
        result = d + val
        self.assertTrue(isinstance(result, Date))

        val = TimeSpan(1, unit="h")
        result = d + val
        self.assertTrue(isinstance(result, DateTimeNano))

    def test_add_datespan(self):
        d = DateSpan(10)
        val = 1
        result = d + val
        self.assertTrue(isinstance(result, DateSpan))

        val = [1]
        result = d + val
        self.assertTrue(isinstance(result, DateSpan))

        val = FastArray(1)
        result = d + val
        self.assertTrue(isinstance(result, DateSpan))

        val = DateSpan(1)
        result = d + val
        self.assertTrue(isinstance(result, DateSpan))

        val = DateTimeNano(NANOS_PER_DAY, from_tz="GMT")
        result = d + val
        self.assertTrue(isinstance(result, DateTimeNano))

        val = TimeSpan(1, unit="h")
        result = d + val
        self.assertTrue(isinstance(result, TimeSpan))

    def test_math_errors(self):

        d = Date(1)
        val = Date(1)
        with self.assertRaises(TypeError):
            result = d + val
        with self.assertRaises(TypeError):
            d += val
        with self.assertRaises(TypeError):
            d -= val

        val = DateTimeNano(NANOS_PER_DAY, from_tz="GMT")
        with self.assertRaises(TypeError):
            result = d + val
        with self.assertRaises(TypeError):
            d += val
        with self.assertRaises(TypeError):
            d -= val

        # with self.assertRaises(TypeError):
        #    result = abs(d)

    def test_hstack(self):
        d = Date(1)
        result = hstack([d, d])
        self.assertTrue(isinstance(result, Date))
        self.assertEqual(len(result), 2)

        d = DateSpan(1)
        result = hstack([d, d])
        self.assertTrue(isinstance(result, DateSpan))
        self.assertEqual(len(result), 2)

    # removed Date, DateSpan from allowed groupby operation types
    # need a better list of computable functions
    # SJK 3/13/2019
    # def test_groupby(self):
    #    ds = Dataset({'col_'+str(i): np.random.rand(5) for i in range(5)})
    #    ds.dates = Date(DateTimeNano.random(5))
    #    ds.datespans = DateSpan(np.random.randint(0,100,5))

    #    result = ds.gb('dates').sum()
    #    self.assertTrue(isinstance(result.dates, Date))
    #    self.assertTrue(isinstance(result.datespans, DateSpan))
    #    self.assertTrue(issorted(result.dates))

    #    result = ds.gb('datespans').sum()
    #    self.assertTrue(isinstance(result.dates, Date))
    #    self.assertTrue(isinstance(result.datespans, DateSpan))
    #    self.assertTrue(issorted(result.datespans))

    def test_comparisons_scalar(self):
        d = Date(10)
        cmp_val = 1
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = 10
        self.assertTrue((d == cmp_val)[0])

        cmp_val = "1970-01-02"
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = "1970-01-11"
        self.assertTrue((d == cmp_val)[0])

        cmp_val = b"1970-01-02"
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = b"1970-01-11"
        self.assertTrue((d == cmp_val)[0])

    def test_comparisons_categorical(self):
        d = Date(10)
        cmp_val = Categorical([b"1970-01-02"])
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = Categorical([b"1970-01-11"])
        self.assertTrue((d == cmp_val)[0])

    def test_comparisons_nano(self):
        d = Date(10)
        cmp_val = DateTimeNano("1970-01-02", from_tz="NYC", to_tz="NYC")
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = DateTimeNano("1970-01-11", from_tz="NYC", to_tz="NYC")
        self.assertTrue((d == cmp_val)[0])

    def test_comparisons_error(self):
        d = Date(10)
        with self.assertRaises(TypeError):
            _ = d > DateSpan(1)
        with self.assertRaises(TypeError):
            _ = d < TimeSpan(1)

    def test_comparisons_datespan_scalar(self):
        d = DateSpan(10)
        cmp_val = 1
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = 10
        self.assertTrue((d == cmp_val)[0])

        cmp_val = 1.0
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = 10.0
        self.assertTrue((d == cmp_val)[0])

    def test_comparisons_datespan_nano(self):
        d = DateSpan(10)
        cmp_val = TimeSpan(24, "h")
        self.assertTrue((d > cmp_val)[0])
        self.assertTrue((d != cmp_val)[0])
        self.assertTrue((d >= cmp_val)[0])
        self.assertFalse((d < cmp_val)[0])
        self.assertFalse((d == cmp_val)[0])
        self.assertFalse((d <= cmp_val)[0])
        cmp_val = TimeSpan(240, "h")
        self.assertTrue((d == cmp_val)[0])

    def test_comparisons_datespan_error(self):
        d = DateSpan(10)
        with self.assertRaises(TypeError):
            _ = d < Date(1)
        with self.assertRaises(TypeError):
            _ = d > DateTimeNano("1970-01-02", from_tz="NYC", to_tz="NYC")
        with self.assertRaises(TypeError):
            _ = d == Categorical(["1970-01-02"])

    def test_year(self):
        arr = full(50, 50) + (arange(50) * 365)
        d = Date(arr)
        result = d.year
        correct = arange(1970, 2020)
        self.assertTrue(bool(np.all(result == correct)))

    def test_month(self):
        d = Date(arange(12) * 32)
        result = d.month
        correct = arange(1, 13)
        self.assertTrue(bool(np.all(result[1:] == correct[1:])))

    def test_day_of_year(self):
        d = Date(arange(365) + 365)
        result = d.day_of_year
        correct = arange(1, 366)
        self.assertTrue(bool(np.all(result == correct)))

    def test_day_of_week(self):
        d = Date.range("2019-02-11", "2019-02-18", closed="left")
        result = d.day_of_week
        correct = arange(7)
        self.assertTrue(bool(np.all(result == correct)))

    def test_is_weekend(self):
        d = Date.range("2019-02-11", "2019-02-18", closed="left")
        result = d.is_weekend
        correct = full(7, False)
        correct[-2:] = True
        self.assertTrue(bool(np.all(result == correct)))

    def test_is_weekday(self):
        d = Date.range("2019-02-11", "2019-02-18", closed="left")
        result = d.is_weekday
        correct = full(7, True)
        correct[-2:] = False
        self.assertTrue(bool(np.all(result == correct)))

    def test_is_leapyear(self):
        d = Date.range("1992-06-01", "2095-06-01", step=365)
        result = d.is_leapyear
        correct = tile([True, False, False, False], 26)
        self.assertTrue(bool(np.all(result == correct)))

    def test_day_of_month(self):
        d = Date.range("2019-01-01", "2020-01-01", closed="left")
        daynums = []
        for idx, i in enumerate(YDAY_SPLITS[:-1]):
            arr = arange(1, YDAY_SPLITS[idx + 1] - i + 1)
            daynums.append(arr)
        daynums.append(arange(1, 32))
        result = d.day_of_month
        correct = hstack(daynums)

        d_leap = Date.range("2020-01-01", "2021-01-01", closed="left")
        daynums = []
        for idx, i in enumerate(YDAY_SPLITS_LEAP[:-1]):
            arr = arange(1, YDAY_SPLITS_LEAP[idx + 1] - i + 1)
            daynums.append(arr)
        daynums.append(arange(1, 32))
        result = d.day_of_month
        correct = hstack(daynums)

    def test_fill_invalid(self):
        dtn = Date.range(20190201, 20190207)
        dtn2 = dtn.fill_invalid(inplace=False)
        self.assertTrue(isinstance(dtn2, Date))
        self.assertTrue(bool(np.all(dtn2.isnan())))

        dtn2 = dtn.fill_invalid(inplace=False, shape=3)
        self.assertTrue(isinstance(dtn2, Date))
        self.assertTrue(bool(np.all(dtn2.isnan())))
        self.assertEqual(len(dtn2), 3)

        with self.assertRaises(ValueError):
            dtn.fill_invalid(inplace=True, shape=3)

        dtn.fill_invalid(inplace=True)
        self.assertTrue(bool(np.all(dtn.isnan())))

    def test_fill_invalid_span(self):
        dtn = DateSpan(arange(5))
        dtn2 = dtn.fill_invalid(inplace=False)
        self.assertTrue(isinstance(dtn2, DateSpan))
        self.assertTrue(bool(np.all(dtn2.isnan())))

        dtn2 = dtn.fill_invalid(inplace=False, shape=3)
        self.assertTrue(isinstance(dtn2, DateSpan))
        self.assertTrue(bool(np.all(dtn2.isnan())))
        self.assertEqual(len(dtn2), 3)

        with self.assertRaises(ValueError):
            dtn.fill_invalid(inplace=True, shape=3)

        dtn.fill_invalid(inplace=True)
        self.assertTrue(bool(np.all(dtn.isnan())))

    def test_min(self):
        d = Date(["20190201", "20190207", "20191230", "20190101"])
        m = d.min()
        self.assertTrue(isinstance(m, Date))
        self.assertTrue(m[0] == d[-1])

        ds = DateSpan([10, 5, 2, 3, 12])
        m = ds.min()
        self.assertTrue(isinstance(m, DateSpan))
        self.assertTrue(m[0] == ds[2])

    def test_max(self):
        d = Date(["20190201", "20190207", "20191230", "20190101"])
        m = d.max()
        self.assertTrue(isinstance(m, Date))
        self.assertTrue(m[0] == d[2])

        ds = DateSpan([10, 5, 2, 3, 12])
        m = ds.max()
        self.assertTrue(isinstance(m, DateSpan))
        self.assertTrue(m[0] == ds[-1])

    def test_range_closed(self):
        both = Date(["20190201", "20190202", "20190203", "20190204"])
        d_both = Date.range("20190201", "20190204")
        self.assertTrue(isinstance(d_both, Date))
        self.assertTrue(bool(np.all(both == d_both)))

        d_left = Date.range("20190201", "20190204", closed="left")
        self.assertTrue(isinstance(d_left, Date))
        self.assertTrue(bool(np.all(both[:-1] == d_left)))

        d_right = Date.range("20190201", "20190204", closed="right")
        self.assertTrue(isinstance(d_right, Date))
        self.assertTrue(bool(np.all(both[1:] == d_right)))

        with self.assertRaises(ValueError):
            d_error = Date.range("20190201", "20190204", closed="garbage")
        pass

    def test_start_of_week(self):
        d = Date(["20200507", "20200508", "20991231"])
        d2 = Date(["20200504", "20200504", "20991228"])
        self.assertTrue((d.start_of_week == d2).all())

    def test_start_of_month(self):
        d = Date(["20200507", "20200508", "20991231"])
        d2 = Date(["20200501", "20200501", "20991201"])
        self.assertTrue((d.start_of_month == d2).all())


if __name__ == "__main__":
    tester = unittest.main()



from setuptools import setup

package_name = "riptable"

install_requires = [
    "ansi2html >=1.5.2",
    "ipykernel",
    "numba >=0.56.2",
    "numpy >=1.22",
    "pandas >=0.24,<2.0",
    "python-dateutil",
    "riptide_cpp >=1.11.2,<2",
]

setup(
    name=package_name,
    packages=[package_name],
    use_scm_version={
        "version_scheme": "post-release",
        "write_to": "riptable/_version.py",
        "write_to_template": '__version__ = "{version}"',
    },
    setup_requires=["setuptools_scm"],
    package_dir={package_name: "riptable"},
    description="Python Package for riptable studies framework",
    author="RTOS Holdings",
    author_email="rtosholdings-bot@sig.com",
    url="https://github.com/rtosholdings/riptable",
    python_requires=">=3.8",
    install_requires=install_requires,
    zip_safe=False,
    include_package_data=True,
    package_data={
        package_name: [
            "tests/*",
            "tests/test_files/*",
            "benchmarks/*",
            "hypothesis_tests/*",
            "Utils/*",
            "test_tooling_integration/*",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)



from .groupby_categorical_unit_test_parameters import *


def generate(n):

    PARAMETERS = categorical_parameters()
    epsilon = 0.00003

    def str_filename():
        return "test_sfw_categorical_autotest_aggregated_functions.py"

    # TODO use relative import for riptable
    def str_imports():
        s = "import pandas as pd"
        s += "\nimport unittest"
        s += "\nimport riptable as rt"
        s += "\n\nfrom .groupby_categorical_unit_test_parameters import *"
        s += "\n\n\ndefault_base_index = 0\n\n"

        return s

    def str_header():
        return """
##                                                                      ##
#                                                                        #
#   THIS TEST WAS AUTOGENERATED BY generator_categorical_unit_test.py    #
#                                                                        #
##                                                                      ##
"""

    def str_agg_params():
        return PARAMETERS.aggs

    def str_cat_kwargs_params():
        return PARAMETERS.kwargs[0]

    def str_pd_kwargs_params():
        return PARAMETERS.kwargs[1]

    def str_agg_params_quoted():
        return '"' + str_agg_params() + '"'

    def str_params():
        return (
            str(PARAMETERS.val_cols)
            + ", "
            + str(PARAMETERS.symbol_ratio + epsilon)[0:4]
            + ", "
            + str_agg_params_quoted()
        )

    def str_name_symratio():
        return "symb_" + str(PARAMETERS.symbol_ratio + epsilon)[0:4].replace(".", "_")

    def str_name_aggs():
        return "aggs_" + str_agg_params()

    def str_name_cols():
        return "ncols_" + str(PARAMETERS.val_cols)

    def str_test_name():
        return str_name_aggs() + "_" + str_name_symratio() + "_" + str_name_cols()

    def str_classname():
        # TODO camel case class names
        return "categorical_test"

    def str_class_header():
        return "\nclass " + str_classname() + "(unittest.TestCase):"

    def str_main_tail():
        return '\n\nif __name__ == "__main__":\n    tester = unittest.main()\n'

    def str_recycle_off():
        return "\n\t\trt.FastArray._ROFF()"

    def str_recycle_on():
        return "\n\t\trt.FastArray._RON()"

    def str_threads_off():
        return "\n\t\trt.FastArray._TOFF()"

    def str_threads_on():
        return "\n\t\trt.FastArray._TON()"

    def str_standard_test():
        s = ""
        s += "\n\n    def test_" + str_test_name() + "(self):"
        s += "\n        test_class = categorical_base(" + str_params() + ")"
        s += (
            "\n        cat = rt.Categorical(\n            values="
            + "test_class.bin_ids,\n            "
            + "categories=test_class.keys,\n            "
            + "base_index=default_base_index,\n        )"
        )
        s += "\n        cat = cat." + str_agg_params() + "(rt.Dataset(test_class.data)" + str_cat_kwargs_params() + ")"
        # s += '\n    ' + 'test_class = rt.Categorical('
        s += "\n        gb = pd.DataFrame(test_class.data)"
        s += "\n        gb = gb.groupby(test_class.bin_ids)." + str_agg_params() + "(" + str_pd_kwargs_params() + ")"

        s += "\n\t\tfor k,v in test_class.data.items():"
        s += "\n\t\t\tsafe_assert(remove_nan(gb[k]), remove_nan(cat[k]))"

        return s

    def str_nothread_test():
        s = ""
        s += "\n\n\tdef test_nothreads_" + str_test_name() + "(self):"
        s += str_threads_off()
        s += "\n\t\ttest_class = categorical_base(" + str_params() + ")"
        s += "\n\t\tcat = rt.Categorical(values=" + "test_class.bin_ids" + ", categories= test_class.keys)"
        s += "\n\t\tcat = cat." + str_agg_params() + "(rt.Dataset(test_class.data))"
        # s += '\n\t' + 'test_class = rt.Categorical('
        s += "\n\t\tgb  = pd.DataFrame(test_class.data)"
        s += "\n\t\tgb  = gb.groupby(test_class.bin_ids)." + str_agg_params() + "()"

        s += "\n\t\tfor k,v in test_class.data.items():"
        s += "\n\t\t\tsafe_assert(remove_nan(gb[k]), remove_nan(cat[k]))"
        s += str_threads_on()

        return s

    def str_norecycle_test():
        s = ""
        s += "\n\n\tdef test_norecycle_" + str_test_name() + "(self):"
        s += str_recycle_off()
        s += "\n\t\ttest_class = categorical_base(" + str_params() + ")"
        s += "\n\t\tcat = rt.Categorical(values=" + "test_class.bin_ids" + ", categories= test_class.keys)"
        s += "\n\t\tcat = cat." + str_agg_params() + "(rt.Dataset(test_class.data))"
        # s += '\n\t' + 'test_class = rt.Categorical('
        s += "\n\t\tgb  = pd.DataFrame(test_class.data)"
        s += "\n\t\tgb  = gb.groupby(test_class.bin_ids)." + str_agg_params() + "()"

        s += "\n\t\tfor k,v in test_class.data.items():"
        s += "\n\t\t\tsafe_assert(remove_nan(gb[k]), remove_nan(cat[k]))"
        s += str_recycle_on()

        return s

    def str_nothreads_norecycle_test():
        s = ""
        s += "\n\n\tdef test_nothreads_norecycle_" + str_test_name() + "(self):"
        s += str_recycle_off()
        s += str_threads_off()
        s += "\n\t\ttest_class = categorical_base(" + str_params() + ")"
        s += "\n\t\tcat = rt.Categorical(values=" + "test_class.bin_ids" + ", categories= test_class.keys)"
        s += "\n\t\tcat = cat." + str_agg_params() + "(rt.Dataset(test_class.data))"
        # s += '\n\t' + 'test_class = rt.Categorical('
        s += "\n\t\tgb  = pd.DataFrame(test_class.data)"
        s += "\n\t\tgb  = gb.groupby(test_class.bin_ids)." + str_agg_params() + "()"

        s += "\n\t\tfor k,v in test_class.data.items():"
        s += "\n\t\t\tsafe_assert(remove_nan(gb[k]), remove_nan(cat[k]))"
        s += str_recycle_on()
        s += str_threads_on()

        return s

    s = "### Filename = " + str_filename() + "\n"
    s += str_header()
    s += str_imports()

    s += str_class_header()

    for tests_numb in range(0, n):
        s += str_standard_test()
        # s += str_norecycle_test()                   ## remove once switched
        # s += str_nothread_test()                    ## remove once switched
        # s += str_nothreads_norecycle_test()         ## remove once switched

        PARAMETERS.update()

    print(s)

    s += str_main_tail()
    filename = str_filename()
    file = open(filename, "w")
    file.write(s)
    file.close()


if __name__ == "__main__":
    generate(35)



"""Unit tests for conversions of arrays and Dataset/tables between riptable and pyarrow."""

import numpy as np
import pytest

import riptable as rt
from riptable.testing.array_assert import assert_array_equal, assert_array_or_cat_equal

# pyarrow is still an optional dependency;
# use the functionality of pytest to skip all tests in this module if pyarrow can't be imported.
# https://docs.pytest.org/en/latest/how-to/skipping.html#skipping-on-a-missing-import-dependency
_ = pytest.importorskip("pyarrow", minversion="4.0.0")
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.types as pat


class TestPyarrowConvertFastArray:
    @pytest.mark.parametrize(
        ("rt_farr",),
        [
            pytest.param(rt.FA([], dtype=np.int8), id="empty(int8)"),
            pytest.param(rt.FA([-120, rt.int8.inv, -1, 0, 1, 101, 127], dtype=np.int8), id="int8"),
            pytest.param(rt.FA([0.01, -0.0, np.nan, 1e10, -1e10, np.inf, np.pi], dtype=np.float32), id="float32"),
            # bool
            pytest.param(rt.FA([b"ABC", b"abcde"], dtype="S"), id="ascii"),
            # FAILS# pytest.param(rt.FA(['ABC', 'abcde'], unicode=True), id='unicode'),
            pytest.param(rt.FA(["A\u1F600C", "abcde"], unicode=True), id="unicode 2"),
        ],
    )
    def test_roundtrip_rt_pa_rt(self, rt_farr: rt.FastArray) -> None:
        """Test round-tripping from rt.FastArray to pyarrow.Array and back."""
        result_pa_arr = rt_farr.to_arrow()
        result_farr = rt.FastArray.from_arrow(result_pa_arr, zero_copy_only=False)
        assert_array_equal(rt_farr, result_farr)

    @pytest.mark.parametrize(
        ("rt_farr",),
        [
            pytest.param(rt.FA([b"ABC", b"abcde"], dtype="S"), id="ascii"),
            pytest.param(rt.FA(["ABC", "abcde"], unicode=True), id="unicode"),
            pytest.param(rt.FA(["A\u1F600C", "abcde"], unicode=True), id="unicode 2"),
        ],
    )
    def test_rt_pa_str(self, rt_farr: rt.FastArray) -> None:
        """Test round-tripping from rt.FastArray to pyarrow.Array and back."""
        result_pa_arr = rt_farr.to_arrow()
        result_list = result_pa_arr.to_pylist()
        assert len(rt_farr) == len(result_list)
        for i in range(len(rt_farr)):
            str_farr = rt_farr[i]
            str_result = result_list[i]
            assert len(str_farr) == len(str_result)
            # TODO: add contents check; frustratingly difficult to do this generically


class TestPyarrowConvertDate:
    @pytest.mark.parametrize(
        ("rt_date_arr",),
        [
            pytest.param(rt.Date([]), id="empty"),
            # TODO: Add test cases
        ],
    )
    def test_roundtrip_rt_pa_rt(self, rt_date_arr: rt.Date) -> None:
        """Test round-tripping from rt.Date to pyarrow.Array and back."""
        result_pa_arr = rt_date_arr.to_arrow()
        result_date_arr = rt.Date.from_arrow(result_pa_arr, zero_copy_only=False)
        assert_array_equal(rt_date_arr, result_date_arr)


class TestPyarrowConvertDateTimeNano:
    @pytest.mark.parametrize(
        ("rt_dtn_arr",),
        [
            pytest.param(rt.DateTimeNano([]), id="empty"),
            # TODO: Add test cases -- including various timezones
        ],
    )
    def test_roundtrip_rt_pa_rt(self, rt_dtn_arr: rt.Date) -> None:
        """Test round-tripping from rt.Date to pyarrow.Array and back."""
        result_pa_arr = rt_dtn_arr.to_arrow()
        result_dtn_arr = rt.DateTimeNano.from_arrow(result_pa_arr, zero_copy_only=False)
        assert_array_equal(rt_dtn_arr, result_dtn_arr)


class TestPyarrowConvertTimeSpan:
    @pytest.mark.parametrize(
        ("rt_tsp_arr",),
        [
            pytest.param(rt.TimeSpan([]), id="empty"),
            # TODO: Add test cases
        ],
    )
    def test_roundtrip_rt_pa_rt(self, rt_tsp_arr: rt.Date) -> None:
        """Test round-tripping from rt.Date to pyarrow.Array and back."""
        result_pa_arr = rt_tsp_arr.to_arrow()
        result_tsp_arr = rt.TimeSpan.from_arrow(result_pa_arr, zero_copy_only=False)
        assert_array_equal(rt_tsp_arr, result_tsp_arr)


class TestPyarrowConvertCategorical:
    @pytest.mark.parametrize(
        ("rt_cat",),
        [
            # TODO: Add test cases for CategoryMode.IntEnum; at present, it appears IntEnum support is broken, can't seem to create a Categorical in that mode.
            # pytest.param(rt.Categorical([]), id='empty', marks=pytest.mark.skip(reason="rt.Categorical does not support creation from an empty list/array.")),
            pytest.param(
                rt.Categorical(["red", "red", "green", "blue", "green", "red", "blue"], ordered=False),
                id="CategoryMode.StringArray",
            ),
            pytest.param(
                rt.Categorical(["red", "red", "green", "blue", "green", "red", "blue"], ordered=True),
                id="CategoryMode.StringArray--ordered",
            ),
            pytest.param(
                rt.Categorical(["red", "red", "green", "blue", "green", "red", "blue"], dtype=np.int8, ordered=False),
                id="CategoryMode.StringArray;int8;ordered=False",
            ),
            pytest.param(
                rt.Categorical(["red", "red", "green", "blue", "green", "red", "blue"], dtype=np.int8, ordered=True),
                id="CategoryMode.StringArray;int8;ordered=True",
            ),
            pytest.param(
                rt.Categorical([f"x{i}" for i in range(0, 127)], dtype=np.int8),
                id="max number of categories for a signed int backing array without causing overflow",
            ),
            # N.B. The test cases below for Categorical[Date] require pyarrow 5.0.0 or higher; dictionary-encoded date32() arrays didn't work before then.
            pytest.param(
                rt.Categorical(
                    rt.Date(
                        [
                            "2019-06-19",
                            "2019-06-19",
                            "2020-01-15",
                            "2020-05-22",
                            "2020-02-10",
                            "2020-02-10",
                            "2020-03-17",
                            "2020-03-17",
                        ]
                    ),
                    ordered=False,
                ),
                id="Categorical[Date];ordered=False",
            ),
            pytest.param(
                rt.Categorical(
                    rt.Date(
                        [
                            "2019-06-19",
                            "2019-06-19",
                            "2020-01-15",
                            "2020-05-22",
                            "2020-02-10",
                            "2020-02-10",
                            "2020-03-17",
                            "2020-03-17",
                        ]
                    ),
                    ordered=True,
                ),
                id="Categorical[Date];ordered=True",
            ),
            pytest.param(
                rt.Categorical(
                    # Country codes -- adapted from TestCategorical.test_hstack_fails_for_different_mode_cats.
                    [36, 36, 344, 840, 840, 372, 840, 372, 840, 124, 840, 124, 36, 484],
                    {
                        "IRL": 372,
                        "USA": 840,
                        "AUS": 36,
                        "HKG": 344,
                        "JPN": 392,
                        "MEX": 484,
                        "KHM": 116,
                        "THA": 764,
                        "JAM": 388,
                        "ARM": 51,
                    },
                    ordered=False,
                ),
                id="CategoryMode.Dictionary;ordered=False;Unicode",
            ),
            pytest.param(
                rt.Categorical(
                    # Country codes -- adapted from TestCategorical.test_hstack_fails_for_different_mode_cats.
                    [36, 36, 344, 840, 840, 372, 840, 372, 840, 124, 840, 124, 36, 484],
                    {
                        "IRL": 372,
                        "USA": 840,
                        "AUS": 36,
                        "HKG": 344,
                        "JPN": 392,
                        "MEX": 484,
                        "KHM": 116,
                        "THA": 764,
                        "JAM": 388,
                        "ARM": 51,
                    },
                    ordered=True,
                ),
                id="CategoryMode.Dictionary;ordered=True;Unicode",
            ),
            pytest.param(
                rt.Categorical(
                    # Country codes -- adapted from TestCategorical.test_hstack_fails_for_different_mode_cats.
                    [36, 36, 344, 840, 840, 372, 840, 372, 840, 124, 840, 124, 36, 484],
                    {
                        b"IRL": 372,
                        b"USA": 840,
                        b"AUS": 36,
                        b"HKG": 344,
                        b"JPN": 392,
                        b"MEX": 484,
                        b"KHM": 116,
                        b"THA": 764,
                        b"JAM": 388,
                        b"ARM": 51,
                    },
                    ordered=False,
                ),
                id="CategoryMode.Dictionary;ordered=False;ASCII",
            ),
            pytest.param(
                rt.Categorical(
                    # Country codes -- adapted from TestCategorical.test_hstack_fails_for_different_mode_cats.
                    [36, 36, 344, 840, 840, 372, 840, 372, 840, 124, 840, 124, 36, 484],
                    {
                        b"IRL": 372,
                        b"USA": 840,
                        b"AUS": 36,
                        b"HKG": 344,
                        b"JPN": 392,
                        b"MEX": 484,
                        b"KHM": 116,
                        b"THA": 764,
                        b"JAM": 388,
                        b"ARM": 51,
                    },
                    ordered=True,
                ),
                id="CategoryMode.Dictionary;ordered=True;ASCII",
            ),
            pytest.param(
                rt.Categorical(
                    [
                        rt.FastArray(
                            ["Cyan", "Magenta", "Yellow", "Black", "Magenta", "Cyan", "Black", "Yellow"]
                        ).set_name("InkColor"),
                        rt.Date(
                            [
                                "2019-06-19",
                                "2019-06-19",
                                "2020-01-15",
                                "2020-05-22",
                                "2020-02-10",
                                "2020-02-10",
                                "2020-03-17",
                                "2020-03-17",
                            ]
                        ).set_name("CartridgeInstallDate"),
                    ]
                ),
                id="CategoryMode.MultiKey",
            ),
        ],
    )
    @pytest.mark.parametrize("output_writable", [False, True])
    @pytest.mark.parametrize("have_nulls", [False, True])
    def test_roundtrip_rt_pa_rt(self, rt_cat: rt.Categorical, output_writable: bool, have_nulls: bool) -> None:
        """Test round-tripping from rt.Categorical to pyarrow.Array/pyarrow.Table and back."""
        orig_cat_shape = rt_cat.shape
        if have_nulls:
            # riptable's filtering/masking uses a valid mask (where False means null/NA).
            indices = np.arange(len(rt_cat))
            valid_mask = indices % 3 != 1
            rt_cat = rt_cat.set_valid(valid_mask)
            assert rt_cat.shape == orig_cat_shape

            # isfiltered() doesn't work as expected for Dictionary/IntEnum-mode Categorical as of riptable 1.1.0.
            filtered_element_count = (
                rt.isnan(rt_cat._fa)
                if rt_cat.category_mode in (rt.rt_enum.CategoryMode.Dictionary, rt.rt_enum.CategoryMode.IntEnum)
                else rt_cat.isfiltered()
            ).sum()
            assert filtered_element_count == (len(rt_cat) - valid_mask.sum())

        result_pa_arr = rt_cat.to_arrow()

        # Verify the pyarrow array has the correct length, number of categories, etc.
        assert len(rt_cat) == len(result_pa_arr)
        assert pat.is_dictionary(result_pa_arr.type)
        assert len(result_pa_arr.dictionary) >= len(
            next(iter(rt_cat.category_dict.values()))
        ), "The number of categories in the pyarrow array's dictionary is smaller than the number of categories in the input Categorical."

        if have_nulls:
            assert valid_mask.sum() > 0
            assert (len(rt_cat) - valid_mask.sum()) == result_pa_arr.null_count

        # TEMP: Certain cases are marked as XFAIL here due to issues in Categorical.
        #         * Cannot create a pre-filtered (i.e. filtered at construction time) Dictionary- or IntEnum-mode Categorical.
        #         * Filtering a Dictionary- or IntEnum-mode Categorical causes unused categories to be dropped,
        #           which is not the same behavior as for other Categorical modes.
        #         * MultiKey Categoricals can't be created with an explicit list of category arrays + an index array,
        #           like what is supported for other Categorical modes.
        if rt_cat.category_mode == rt.rt_enum.CategoryMode.MultiKey or (
            have_nulls and rt_cat.category_mode == rt.rt_enum.CategoryMode.Dictionary
        ):
            pytest.xfail("Expected failure due to issues with the Categorical constructor and/or filtering.")

        result_cat = rt.Categorical.from_arrow(result_pa_arr, zero_copy_only=False, writable=output_writable)

        # relaxed_cat_check <==> rt_cat.ordered, because if the categories are ordered, we expect them to be
        # in the same position after being roundtripped, so they should be mapped to the same integer before/after.
        # multi-key cats always seem to be ordered, even if ordered=False is specified when creating them.
        # TODO: Remove CategoryMode.Dictionary from the relaxed_cat_check here -- it's failing because our encoding in
        #       pyarrow doesn't currenly preserve unused entries from the name <-> code mapping. Once that's fixed
        #       we should be able to use the stronger equality check.
        assert_array_or_cat_equal(
            rt_cat,
            result_cat,
            relaxed_cat_check=rt_cat.ordered
            or rt_cat.category_mode == rt.rt_enum.CategoryMode.MultiKey
            or rt_cat.category_mode == rt.rt_enum.CategoryMode.Dictionary,
        )

    @pytest.mark.parametrize(
        ("num_cats", "dtype"),
        [
            pytest.param(127, np.uint8),
            pytest.param(128, np.uint8),
            pytest.param(129, np.uint8),
            pytest.param(32769, np.uint16),
        ],
    )
    @pytest.mark.parametrize("ordered", [False, True])
    @pytest.mark.parametrize("output_writable", [False, True])
    @pytest.mark.parametrize("have_nulls", [False, True])
    def test_pa_to_rt_unsigned(self, num_cats, dtype, ordered: bool, output_writable: bool, have_nulls: bool) -> None:
        # Create a numpy array containing `num_cats` distinct strings.
        cat_labels = np.array([f"x{i}" for i in range(0, num_cats)])
        indices = np.arange(num_cats, dtype=dtype)

        # Create the pyarrow dict-encoded array.
        if have_nulls:
            # pyarrow uses an INvalid mask (where True means null/NA).
            invalid_mask = indices % 7 == 3
            pa_indices = pa.array(indices, mask=invalid_mask)
            pa_arr = pa.DictionaryArray.from_arrays(pa_indices, cat_labels, ordered=ordered)
        else:
            pa_arr = pa.DictionaryArray.from_arrays(indices, cat_labels, ordered=ordered)

        assert len(pa_arr.dictionary) == num_cats

        # Create the Categorical from the pyarrow array.
        result_cat = rt.Categorical.from_arrow(pa_arr, zero_copy_only=False, writable=output_writable)

        if have_nulls:
            result_invalid_mask = result_cat.isfiltered()
            assert_array_equal(result_invalid_mask, invalid_mask)

        # TODO: Add assertions here to verify correctness of `result_cat`.
        #   * Make sure we have same number of categories as input pyarrow arrray, even if not all categories (entries in the pyarrow array.dictionary) are used.


class TestPyarrowConvertDataset:
    @pytest.mark.parametrize(
        ("rt_dset",),
        [
            pytest.param(rt.Dataset({}), id="empty"),
            pytest.param(
                rt.Dataset(
                    {
                        "ink_capacity": rt.FA([15, 10, 15, 25, 10, 15, 25, 15]),
                        "purchase_date": rt.Date(
                            [
                                "2019-06-19",
                                "2019-06-19",
                                "2020-01-15",
                                "2020-05-22",
                                "2020-02-10",
                                "2020-02-10",
                                "2020-03-17",
                                "2020-03-17",
                            ]
                        ),
                        "country_code": rt.Categorical(
                            # Country codes -- adapted from TestCategorical.test_hstack_fails_for_different_mode_cats.
                            [36, 36, 344, 840, 840, 124, 36, 484],
                            {
                                "IRL": 372,
                                "USA": 840,
                                "AUS": 36,
                                "HKG": 344,
                                "JPN": 392,
                                "MEX": 484,
                                "KHM": 116,
                                "THA": 764,
                                "JAM": 388,
                                "ARM": 51,
                            },
                            ordered=True,
                        ),
                    }
                )
            ),
        ],
    )
    def test_roundtrip_rt_pa_rt(self, rt_dset: rt.Dataset) -> None:
        """Test round-tripping from rt.Dataset to pyarrow.Table and back."""
        result_pa_tbl = rt_dset.to_arrow()
        result_rt_dset = rt.Dataset.from_arrow(result_pa_tbl, zero_copy_only=False)

        assert rt_dset.keys() == result_rt_dset.keys()
        for col_name in rt_dset.keys():
            # relaxed_cat_check=True, because we're not trying to test specific details of Categorical conversion
            # here, we're more interested in the dataset-level stuff.
            assert_array_or_cat_equal(rt_dset[col_name], result_rt_dset[col_name], relaxed_cat_check=True)


class TestPyarrowConvertFixedSizeBinary:
    @pytest.mark.parametrize(
        ("pa_arr", "zero_copy", "writable"),
        [
            (pa.array([b"123", b"456"], type=pa.binary(3)), True, False),
            (pa.array([b"123", b"456"], type=pa.binary(3)), False, False),
            (pa.array([b"123", b"456"], type=pa.binary(3)), True, True),
            (pa.array([b"123", b"456"], type=pa.binary(3)), False, False),
            (pa.array([b"123", b"456"], mask=[True, False], type=pa.binary(3)), False, False),
            (pa.array([b"123", b"456"], mask=[True, False], type=pa.binary(3)), False, True),
        ],
    )
    def test_roundtrip_pa_fa_pa(self, pa_arr: pa.Array, zero_copy: bool, writable: bool):
        result_fa_array = rt.FastArray.from_arrow(pa_arr, zero_copy_only=zero_copy, writable=writable)
        result_pa_array = result_fa_array.to_arrow(type=pa.binary(3), preserve_fixed_bytes=True)

        assert pa_arr == result_pa_array
